# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

__name__ = "file-config"
__repo__ = "https://github.com/stephen-bunn/file-config"
__version__ = "0.0.7"
__description__ = "An attr's based configuration file abstraction"
__author__ = "Stephen Bunn"
__contact__ = "stephen@bunn.io"
__license__ = "MIT License"
